use crate::app::AppArgs;
use std::path::PathBuf;

#[derive(Clone)]
pub enum AppContext {
    Debug,
    Production,
}

#[derive(Clone)]
pub struct AppConfig {
    pub exec_path: PathBuf,
    pub exec_name: String,
    pub context: AppContext,
    pub storage_dir: PathBuf,
    pub host: String,
    pub port: u16,
    pub port_tls: u16,
    pub load: Option<PathBuf>,
    pub ssl_certs_dir: PathBuf,
    pub admin_username: String,
    pub admin_pwd: String,
    pub admin_cookie_name: String,
}

impl AppConfig {
    pub fn new(app_args: AppArgs) -> Self {
        let exec_path = std::env::current_exe().unwrap();
        let exec_name = exec_path
            .file_name()
            .expect("Unable to get executable name")
            .to_str()
            .unwrap()
            .to_string();

        let context = Self::get_context(&app_args);
        let storage_dir = Self::get_storage_dir(&app_args, &exec_name, &context).unwrap();
        let ssl_certs_dir = app_args.ssl_certs_dir.join(&app_args.host);

        AppConfig {
            exec_path: std::path::PathBuf::from(exec_path),
            exec_name,
            context,
            storage_dir,
            host: app_args.host,
            port: app_args.port,
            port_tls: app_args.port_tls,
            load: app_args.load,
            ssl_certs_dir,
            admin_username: app_args.admin_username,
            admin_pwd: app_args.admin_pwd,
            admin_cookie_name: app_args.admin_cookie_name,
        }
    }

    fn get_storage_dir(
        app_args: &AppArgs,
        exec_name: &String,
        context: &AppContext,
    ) -> Result<PathBuf, String> {
        match &app_args.app_storage_root {
            Some(dir) => Ok(dir.join(&exec_name)),
            None => match context {
                AppContext::Production => Ok(PathBuf::from("/var").join(&exec_name)),
                AppContext::Debug => {
                    let home_dir = dirs::home_dir();
                    if home_dir.is_none() {
                        return Err(String::from("Unable to get home dir"));
                    }
                    Ok(home_dir.unwrap().join(&exec_name))
                }
            },
        }
    }

    fn get_context(app_args: &AppArgs) -> AppContext {
        match app_args.context.as_str() {
            "prod" => AppContext::Production,
            _ => AppContext::Debug,
        }
    }

    pub fn get_log_level(&self) -> String {
        match self.context {
            AppContext::Debug => "debug".to_string(),
            AppContext::Production => "info".to_string(),
        }
    }
}
